@@ -6,7 +6,7 @@
 ;;          Thierry Volpiatto <thievol@posteo.net>
 
 ;; Keywords: dired async network
-;; X-URL: https://github.com/jwiegley/dired-async
+;; X-URL: https://github.com/jwiegley/emacs-async
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -68,18 +68,24 @@ Should take same args as `message'."
   :risky t
   :type 'sexp)
 
-(defcustom dired-async-skip-fast t
+(defcustom dired-async-skip-fast nil
   "If non-nil, skip async for fast operations.
 Same device renames and copying and renaming files smaller than
-`dired-async-large-file' are considered fast."
+`dired-async-small-file-max' are considered fast.
+If the total size of all files exceed `dired-async-small-file-max'
+operation is not considered fast."
   :risky t
-  :type 'bool)
+  :type 'boolean)
 
 (defcustom dired-async-small-file-max 5000000
   "Files smaller than this in bytes are considered fast to copy
 or rename for `dired-async-skip-fast'."
   :risky t
-  :type 'int)
+  :type 'integer)
+
+(defcustom dired-async-large-file-warning-threshold large-file-warning-threshold
+  "Same as `large-file-warning-threshold' but for dired-async."
+  :type 'integer)
 
 (defface dired-async-message
     '((t (:foreground "yellow")))
@@ -115,10 +121,9 @@ or rename for `dired-async-skip-fast'."
     (sit-for 3)
     (force-mode-line-update)))
 
-(defun dired-async-processes ()
+(defun dired-async-processes (&optional propname)
   (cl-loop for p in (process-list)
-           when (cl-loop for c in (process-command p) thereis
-                         (string= "async-batch-invoke" c))
+           when (process-get p (or propname 'dired-async-process))
            collect p))
 
 (defun dired-async-kill-process ()
@@ -195,42 +200,69 @@ See `file-attributes'."
 
 (defsubst dired-async--same-device-p (f1 f2)
   "Return non-nil if F1 and F2 have the same device number."
-  (= (file-attribute-device-number (file-attributes f1))
-     (file-attribute-device-number (file-attributes f2))))
+  ;; file-attribute-device-number may be a cons cell, so use equal for
+  ;; testing (See Emacs bug/58446).
+  (equal (file-attribute-device-number (file-attributes f1))
+         (file-attribute-device-number (file-attributes f2))))
 
-(defun dired-async--small-file-p (file)
-  "Return non-nil if FILE is small (can create quickly)."
-  (let ((a (file-attributes file)))
+(defun dired-async--small-file-p (file &optional attrs)
+  "Return non-nil if FILE is considered small.
+
+File is considered small if it size is smaller than
+`dired-async-small-file-max'."
+  (let ((a (or attrs (file-attributes file))))
     ;; Directories are always large since we can't easily figure out
     ;; their total size.
     (and (not (dired-async--directory-p a))
-         ;; 5 MB
          (< (file-attribute-size a) dired-async-small-file-max))))
 
-(defun dired-async--skip-async-p (file-creator file name-constructor)
+(defun dired-async--skip-async-p (file-creator file name-constructor &optional attrs)
   "Return non-nil if we should skip async for FILE.
 See `dired-create-files' for FILE-CREATOR and NAME-CONSTRUCTOR."
   ;; Skip async for small files.
-  (or (dired-async--small-file-p file)
+  (or (dired-async--small-file-p file attrs)
       ;; Also skip async for same device renames.
       (and (eq file-creator 'dired-rename-file)
            (let ((new (funcall name-constructor file)))
              (dired-async--same-device-p file (file-name-directory new))))))
 
-(defun dired-async--smart-create-files (old-func file-creator operation fn-list name-constructor
-                                                 &optional marker-char)
+(defun dired-async--smart-create-files (old-func file-creator
+                                        operation fn-list name-constructor
+                                        &optional marker-char)
   "Around advice for `dired-create-files'.
 Uses async like `dired-async-create-files' but skips certain fast
 cases if `dired-async-skip-fast' is non-nil."
-  (let (async-list quick-list)
-    (dolist (old fn-list)
-      (if (dired-async--skip-async-p file-creator old name-constructor)
-          (push old quick-list)
-        (push old async-list)))
+  (let ((total-size 0)
+        async-list quick-list)
+    (if (or (eq file-creator 'backup-file)
+            (null dired-async-skip-fast))
+        (setq async-list fn-list)
+      (dolist (old fn-list)
+        (let ((attrs (file-attributes old)))
+          (if (dired-async--skip-async-p
+               file-creator old name-constructor attrs)
+              (progn
+                (push old quick-list)
+                (setq total-size (+ total-size (nth 7 attrs))))
+            (push old async-list)))))
+    (when (> total-size dired-async-small-file-max)
+      (setq async-list (append quick-list async-list)
+            quick-list nil))
     (when async-list
-      (dired-async-create-files file-creator operation (nreverse async-list) name-constructor marker-char))
+      (dired-async-create-files
+       file-creator operation (nreverse async-list)
+       name-constructor marker-char))
     (when quick-list
-      (funcall old-func file-creator operation (nreverse quick-list) name-constructor marker-char))))
+      (funcall old-func file-creator operation
+               (nreverse quick-list) name-constructor marker-char))))
+
+(defun dired-async--abort-if-file-too-large (size op-type filename)
+  "Warn when FILENAME larger than `dired-async-large-file-warning-threshold'.
+Same as `abort-if-file-too-large' but without user-error."
+  (when (and dired-async-large-file-warning-threshold size
+	     (> size dired-async-large-file-warning-threshold))
+    (files--ask-user-about-large-file
+     size op-type filename nil)))
 
 (defvar overwrite-query)
 (defun dired-async-create-files (file-creator operation fn-list name-constructor
@@ -241,7 +273,7 @@ See `dired-create-files' for the behavior of arguments."
   (setq overwrite-query nil)
   (let ((total (length fn-list))
         failures async-fn-list skipped callback
-        async-quiet-switch)
+        async-quiet-switch create-dir)
     (let (to)
       (dolist (from fn-list)
         (setq to (funcall name-constructor from))
@@ -289,14 +321,22 @@ ESC or `q' to not overwrite any of the remaining files,
                    (file-in-directory-p destname from)
                    (error "Cannot copy `%s' into its subdirectory `%s'"
                           from to)))
-            (if overwrite
-                (or (and dired-overwrite-confirmed
-                         (push (cons from to) async-fn-list))
-                    (progn
-                      (push (dired-make-relative from) failures)
-                      (dired-log "%s `%s' to `%s' failed\n"
-                                 operation from to)))
-              (push (cons from to) async-fn-list)))))
+            ;; Skip file if it is too large.
+            (if (and (member operation '("Copy" "Rename"))
+                     (eq (dired-async--abort-if-file-too-large
+                          (file-attribute-size
+                           (file-attributes (file-truename from)))
+                          (downcase operation) from)
+                         'abort))
+                (push from skipped)
+              (if overwrite
+                  (or (and dired-overwrite-confirmed
+                           (push (cons from to) async-fn-list))
+                      (progn
+                        (push (dired-make-relative from) failures)
+                        (dired-log "%s `%s' to `%s' failed\n"
+                                   operation from to)))
+                (push (cons from to) async-fn-list))))))
       ;; Fix tramp issue #80 with emacs-26, use "-q" only when needed.
       (setq async-quiet-switch
             (if (and (boundp 'tramp-cache-read-persistent-data)
@@ -334,50 +374,67 @@ ESC or `q' to not overwrite any of the remaining files,
                          for destp = (file-exists-p to)
                          do (and bf destp
                                  (with-current-buffer bf
-                                   (set-visited-file-name to t t))))))))
+                                   (set-visited-file-name to t t)))))))
+      (let ((dirp (file-directory-p to))
+            (dest (file-name-directory to)))
+        (when (boundp 'dired-create-destination-dirs)
+          (setq create-dir
+                (cl-case dired-create-destination-dirs
+                  (always 'always)
+                  (ask (and (null dirp)
+                            (null (file-directory-p dest))
+                            (y-or-n-p (format "Create directory `%s'? " dest)))
+                       'always))))))
     ;; Start async process.
     (when async-fn-list
-      (async-start `(lambda ()
-                      (require 'cl-lib) (require 'dired-aux) (require 'dired-x)
-                      ,(async-inject-variables dired-async-env-variables-regexp)
-                      (let ((dired-recursive-copies (quote always))
-                            (dired-copy-preserve-time
-                             ,dired-copy-preserve-time))
-                        (setq overwrite-backup-query nil)
-                        ;; Inline `backup-file' as long as it is not
-                        ;; available in emacs.
-                        (defalias 'backup-file
-                          ;; Same feature as "cp -f --backup=numbered from to"
-                          ;; Symlinks are copied as file from source unlike
-                          ;; `dired-copy-file' which is same as cp -d.
-                          ;; Directories are omitted.
-                          (lambda (from to ok)
-                            (cond ((file-directory-p from) (ignore))
-                                  (t (let ((count 0))
-                                       (while (let ((attrs (file-attributes to)))
-                                                (and attrs (null (nth 0 attrs))))
-                                         (cl-incf count)
-                                         (setq to (concat (file-name-sans-versions to)
-                                                          (format ".~%s~" count)))))
-                                     (condition-case err
-                                         (copy-file from to ok dired-copy-preserve-time)
-                                       (file-date-error
-                                        (dired-log "Can't set date on %s:\n%s\n" from err)))))))
-                        ;; Now run the FILE-CREATOR function on files.
-                        (cl-loop with fn = (quote ,file-creator)
-                                 for (from . dest) in (quote ,async-fn-list)
-                                 do (condition-case err
-                                        (funcall fn from dest t)
-                                      (file-error
-                                       (dired-log "%s: %s\n" (car err) (cdr err))
-                                       nil)))
-                        (when (get-buffer dired-log-buffer)
-                          (dired-log t)
-                          (with-current-buffer dired-log-buffer
-                            (write-region (point-min) (point-max)
-                                          ,dired-async-log-file))))
-                      ,(dired-async-maybe-kill-ftp))
-                   callback)
+      (process-put
+       (async-start `(lambda ()
+                       (require 'cl-lib) (require 'dired-aux) (require 'dired-x)
+                       ,(async-inject-variables dired-async-env-variables-regexp)
+                       (advice-add #'files--ask-user-about-large-file
+                                   :override (lambda (&rest args) nil))
+                       (let ((dired-recursive-copies (quote always))
+                             (dired-copy-preserve-time
+                              ,dired-copy-preserve-time)
+                             (dired-create-destination-dirs ',create-dir)
+                             (dired-vc-rename-file ,dired-vc-rename-file)
+                             auth-source-save-behavior)
+                         (setq overwrite-backup-query nil)
+                         ;; Inline `backup-file' as long as it is not
+                         ;; available in emacs.
+                         (defalias 'backup-file
+                           ;; Same feature as "cp -f --backup=numbered from to"
+                           ;; Symlinks are copied as file from source unlike
+                           ;; `dired-copy-file' which is same as cp -d.
+                           ;; Directories are omitted.
+                           (lambda (from to ok)
+                             (cond ((file-directory-p from) (ignore))
+                                   (t (let ((count 0))
+                                        (while (let ((attrs (file-attributes to)))
+                                                 (and attrs (null (nth 0 attrs))))
+                                          (cl-incf count)
+                                          (setq to (concat (file-name-sans-versions to)
+                                                           (format ".~%s~" count)))))
+                                      (condition-case err
+                                          (copy-file from to ok dired-copy-preserve-time)
+                                        (file-date-error
+                                         (dired-log "Can't set date on %s:\n%s\n" from err)))))))
+                         ;; Now run the FILE-CREATOR function on files.
+                         (cl-loop with fn = (quote ,file-creator)
+                                  for (from . dest) in (quote ,async-fn-list)
+                                  do (condition-case err
+                                         (funcall fn from dest t)
+                                       (file-error
+                                        (dired-log "%s: %s\n" (car err) (cdr err))
+                                        nil)))
+                         (when (get-buffer dired-log-buffer)
+                           (dired-log t)
+                           (with-current-buffer dired-log-buffer
+                             (write-region (point-min) (point-max)
+                                           ,dired-async-log-file))))
+                       ,(dired-async-maybe-kill-ftp))
+                    callback)
+       'dired-async-process t)
       ;; Run mode-line notifications while process running.
       (dired-async--modeline-mode 1)
       (message "%s proceeding asynchronously..." operation))))
